# Copyright (C) 2006-2007 Red Hat, Inc.
#
# This program is free software; you can redistribute
# it and/or modify it under the terms of version 2 of the
# GNU General Public License as published by the
# Free Software Foundation.

import xml
import xml.dom
from xml.dom import minidom
from LuciSyslog import get_logger
from conga_ssl import SSLSocket
from conga_constants import LUCI_DEBUG_MODE, LUCI_DEBUG_NET

CERTS_DIR_PATH = '/var/lib/luci/var/certs/'
luci_log = get_logger()

class RicciError(Exception):
	pass

class RicciCommunicator:
	def __init__(self, hostname, enforce_trust=True, port=11111):
		self.__hostname = hostname
		self.__port = port

		self.__timeout_init  = 4
		self.__timeout_auth  = 4
		self.__timeout_short = 6
		self.__timeout_long  = 600

		self.__privkey_file = '%sprivkey.pem' % CERTS_DIR_PATH
		self.__cert_file = '%scacert.pem' % CERTS_DIR_PATH

		try:
			self.ss = SSLSocket(self.__hostname,
								self.__port,
								self.__timeout_init)
			if enforce_trust:
				if not self.ss.trusted():
					luci_log.info('The SSL certificate for host "%s" is not trusted. Aborting connection attempt.' % self.__hostname)
					raise RicciError, 'ricci\'s certificate is not trusted'
		except Exception, e:
			errmsg = 'Unable to establish an SSL connection to %s:%d: %s' \
						% ((self.__hostname, self.__port, str(e)))
			luci_log.info(errmsg)
			raise RicciError, errmsg

		# receive ricci header
		hello = self.__receive(self.__timeout_init)
		try:
			if LUCI_DEBUG_NET is True:
				luci_log.debug_net('RC:init0: recv header from %s: \"%s\"' \
					% (self.__hostname, hello.toxml()))
		except:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('RC:init0: error receiving header from %s:%d: %r %s' % (self.__hostname, self.__port, e, str(e)))
			else:
				luci_log.info('Error reading from %s:%d: %s' \
					% (self.__hostname, self.__port, str(e)))

		self.__authed = hello.firstChild.getAttribute('authenticated') == 'true'
		self.__cluname = hello.firstChild.getAttribute('clustername')
		self.__clualias = hello.firstChild.getAttribute('clusteralias')
		self.__reported_hostname = hello.firstChild.getAttribute('hostname')
		self.__os = hello.firstChild.getAttribute('os')
		self.__dom0 = hello.firstChild.getAttribute('xen_host') == 'true'

	def hostname(self):
		if LUCI_DEBUG_NET is True:
			luci_log.debug_net('RC:hostname: [auth %d] hostname = %s' \
				% (self.__authed, self.__hostname))
		return self.__hostname

	def authed(self):
		if LUCI_DEBUG_NET is True:
			luci_log.debug_net('RC:authed: reported authed = %d for %s' \
				% (self.__authed, self.__hostname))
		return self.__authed

	def system_name(self):
		if LUCI_DEBUG_NET is True:
			luci_log.debug_net('RC:SN: [auth %d] system_name = %s for %s' \
				% (self.__authed, self.__reported_hostname, self.__hostname))
		return self.__reported_hostname

	def cluster_info(self):
		if LUCI_DEBUG_NET is True:
			luci_log.debug_net('RC:CI: [auth %d] (%s,%s) for %s' \
				% (	self.__authed, self.__cluname,
					self.__clualias, self.__hostname))

		return (self.__cluname, self.__clualias)

	def os(self):
		if LUCI_DEBUG_NET is True:
			luci_log.debug_net('RC:OS: [auth %d] reported os = %s for %s' \
				% (self.__authed, self.__os, self.__hostname))
		return self.__os

	def dom0(self):
		if LUCI_DEBUG_NET is True:
			luci_log.debug_net('RC:DOM0: [auth %d] dom0 = %s for %s' \
				% (self.__authed, self.__dom0, self.__hostname))
		return self.__dom0

	def fingerprint(self):
		fp = self.ss.peer_fingerprint()
		if LUCI_DEBUG_NET is True:
			luci_log.debug_net('RC:fp: [auth %d] fp for %s = %s' \
				% (self.__authed, self.__hostname, fp))
		return fp

	def trust(self):
		return self.ss.trust()

	def untrust(self):
		return self.ss.untrust()

	def trusted(self):
		return self.ss.trusted()

	def auth(self, password):
		if self.authed():
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('RC:AUTH0: already authenticated to %s' \
					% self.__hostname)
			return True

		# send request
		doc = minidom.Document()
		ricci = doc.createElement("ricci")
		ricci.setAttribute("version", "1.0")
		ricci.setAttribute("function", "authenticate")
		ricci.setAttribute("password", password)
		doc.appendChild(ricci)
		self.__send(doc, self.__timeout_auth)

		# receive response
		resp = self.__receive(self.__timeout_auth)
		self.__authed = resp.firstChild.getAttribute('authenticated') == 'true'

		if self.__authed:
			try:
				self.__cluname = resp.firstChild.getAttribute('clustername')
				self.__clualias = resp.firstChild.getAttribute('clusteralias')
				self.__reported_hostname = resp.firstChild.getAttribute('hostname')
				self.__os = resp.firstChild.getAttribute('os')
				self.__dom0 = resp.firstChild.getAttribute('xen_host') == 'true'
			except Exception, e:
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('RC:AUTH1: %r %s for %s' \
						% (e, str(e), self.__hostname))

		if LUCI_DEBUG_NET is True:
			luci_log.debug_net('RC:auth1: auth call returning %d' \
				% self.__authed)
		return self.__authed

	def unauth(self):
		doc = minidom.Document()
		ricci = doc.createElement('ricci')
		ricci.setAttribute('version', '1.0')
		ricci.setAttribute('function', 'unauthenticate')
		doc.appendChild(ricci)
		self.__send(doc, self.__timeout_auth)
		resp = self.__receive(self.__timeout_auth)

		try:
			ret = resp.firstChild.getAttribute('success')
			if LUCI_DEBUG_NET is True:
				luci_log.debug_net('RC:UNAUTH0: unauthenticate %s for %s' \
					% (ret, self.__hostname))
			if ret != '0':
				raise Exception, 'Invalid response'

			try:
				self.ss.untrust()
			except:
				pass
		except:
			errstr = 'Error authenticating host %s: %s' \
						% (self.__hostname, str(ret))
			luci_log.info(errstr)
			raise RicciError, errstr
		return True

	def process_batch(self, batch_xml, async=False):
		try:
			if LUCI_DEBUG_NET is True:
				luci_log.debug_net_priv('RC:PB0: [auth=%d] to %s for batch %s [async=%d]' % (self.__authed, self.__hostname, batch_xml.toxml(), async))
		except:
			pass

		if not self.authed():
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('RC:PB1: not authenticated to %s' \
					% self.__hostname)
			raise RicciError, 'not authenticated to host %s' % self.__hostname

		# construct request
		doc = minidom.Document()
		ricci = doc.createElement('ricci')
		ricci.setAttribute('version', '1.0')
		ricci.setAttribute('function', 'process_batch')
		async_str = None
		if async:
			async_str = 'true'
		else:
			async_str = 'false'
		ricci.setAttribute('async', async_str)
		doc.appendChild(ricci)
		ricci.appendChild(batch_xml)#.cloneNode(True))

		# send request
		try:
			self.__send(doc, self.__timeout_short)
		except Exception, e:
			errstr = 'Error sending XML batch command to %s:%d: %s' \
						% (self.__hostname, self.__port, str(e))
			luci_log.info(errstr)
			if LUCI_DEBUG_NET is True:
				luci_log.debug_net_priv('RC:PB2: Error sending XML \"%s\" to host %s' % (doc.toxml(), self.__hostname))
			raise RicciError, errstr

		# receive response
		doc = self.__receive(self.__timeout_long)
		if LUCI_DEBUG_NET is True:
			try:
				luci_log.debug_net_priv('RC:PB3: received from %s XML \"%s\"' \
					% (self.__hostname, doc.toxml()))
			except:
				pass

		if doc.firstChild.getAttribute('success') != '0':
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('RC:PB4: batch command failed')
			raise RicciError, 'The last ricci command to host %s:%d failed' \
					% (self.__hostname, self.__port)

		batch_node = None
		for node in doc.firstChild.childNodes:
			if node.nodeType == xml.dom.Node.ELEMENT_NODE:
				if node.nodeName == 'batch':
					batch_node = node#.cloneNode(True)
		if batch_node is None:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('RC:PB5: no <batch/> from %s' \
					% self.__hostname)
			raise RicciError, 'missing <batch/> in ricci response from %s:%d' \
					% (self.__hostname, self.__port)

		return batch_node

	def batch_run(self, batch_str, async=True):
		try:
			batch_xml_str = '<?xml version="1.0" ?><batch>%s</batch>' \
				% batch_str
			if LUCI_DEBUG_NET is True:
				luci_log.debug_net_priv('RC:BRun0: attempting batch "%s" for host "%s"' % (batch_xml_str, self.__hostname))
			batch_xml = minidom.parseString(batch_xml_str).firstChild
		except Exception, e:
			if LUCI_DEBUG_NET is True:
				luci_log.debug_net_priv('RC:BRun1: received invalid batch XML for %s: "%s": "%s"' % (self.__hostname, batch_xml_str, str(e)))
			raise RicciError, 'batch XML is malformed'

		try:
			ricci_xml = self.process_batch(batch_xml, async)
			if LUCI_DEBUG_NET is True:
				try:
					luci_log.debug_net_priv('RC:BRun2: received XML "%s" from host %s in response to batch command' % (ricci_xml.toxml(), self.__hostname))
				except:
					pass
		except:
			if LUCI_DEBUG_NET is True:
				luci_log.debug_net_priv('RC:BRun3: An error occurred while trying to process the batch job: "%s"' % batch_xml_str)
			return None

		doc = minidom.Document()
		doc.appendChild(ricci_xml)
		return doc

	def batch_status(self, batch_id):
		err_msg = 'An unknown ricci error occurred on %s:%d' \
					% (self.__hostname, self.__port)

		try:
			batch = self.batch_report(batch_id)
			if batch is None:
				if LUCI_DEBUG_MODE is True:
					luci_log.debug_verbose('RCCB0: batch id %s not found' \
						% batch_id)
				return (True, 'batch id %s was not found' % batch_id)
		except Exception, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('RCCB1: %s:%d: %r %s' \
					% (self.__hostname, self.__port, e, str(e)))
			return (-1, err_msg)

		try:
			code, new_err_msg = extract_module_status(batch, 1)
			if new_err_msg:
				err_msg = 'A ricci error occurred on %s:%d: %s' \
							% (self.__hostname, self.__port, str(new_err_msg))
		except Exception, e:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('RCCB2: %s:%d %r %s: %s' \
					% (self.__hostname, self.__port, e, str(e), err_msg))
			return (-1, err_msg)

		# In progress.
		if code == -101 or code == -102:
			return (False, 'in progress')

		# Done successfully.
		if code == 0:
			return (True, 'completed sucessfully')

		# Error
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('RCCB3: %s:%d code %d: %s' \
				% (self.__hostname, self.__port, code, err_msg))
		return (-1, err_msg)

	def batch_report(self, batch_id):
		if LUCI_DEBUG_NET is True:
			luci_log.debug_net('RC:BRep0: [auth=%d] asking for batchid# %s for host %s' % (self.__authed, batch_id, self.__hostname))

		if not self.authed():
			raise RicciError, 'Not authenticated to host %s' % self.__hostname

		# construct request
		doc = minidom.Document()
		ricci = doc.createElement("ricci")
		ricci.setAttribute("version", "1.0")
		ricci.setAttribute("function", "batch_report")
		ricci.setAttribute("batch_id", str(batch_id))
		doc.appendChild(ricci)

		# send request
		self.__send(doc, self.__timeout_short)

		# receive response
		doc = self.__receive(self.__timeout_short)
		if doc.firstChild.getAttribute('success') == '12':
			return None
		if doc.firstChild.getAttribute('success') != '0':
			raise RicciError, 'Error while retrieving batch report for batch #%d from host %s' % (batch_id, self.__hostname)
		batch_node = None
		for node in doc.firstChild.childNodes:
			if node.nodeType == xml.dom.Node.ELEMENT_NODE:
				if node.nodeName == 'batch':
					batch_node = node#.cloneNode(True)
		if batch_node is None:
			raise RicciError, 'Missing <batch/> in ricci\'s response from host %s' % self.__hostname
		return batch_node


	def __send(self, xml_doc, timeout):
		try:
			self.ss.send('%s\n' % xml_doc.toxml(), timeout)
		except Exception, e:
			if LUCI_DEBUG_NET is True:
				luci_log.debug_net_priv('RC:send0: sending XML "%s" to %s: %s' \
					% (xml_doc.toxml(), self.__hostname, str(e)))
			luci_log.info('Error sending batch command to %s:%d: %s' \
				% (self.__hostname, self.__port, str(e)))
			raise RicciError, 'Write error while sending XML to host %s' \
					% self.__hostname

		if LUCI_DEBUG_NET is True:
			try:
				luci_log.debug_net_priv('RC:send1: Sent XML "%s" to host %s' \
					% (xml_doc.toxml(), self.__hostname))
			except:
				pass
		return

	def __receive(self, timeout):
		xml_in = ''
		try:
			xml_in = self.ss.recv(timeout)
		except Exception, e:
			errstr = 'Error reading from %s:%d: %s' \
						% (self.__hostname, self.__port, str(e))
			luci_log.info(errstr)
			raise RicciError, errstr

		if LUCI_DEBUG_NET is True:
			luci_log.debug_net_priv('RC:recv1: Received XML "%s" from host %s' \
				% (xml_in, self.__hostname))

		doc = None
		try:
			doc = minidom.parseString(xml_in)
		except Exception, e:
			errstr = 'Error parsing XML from host %s:%d: %s' \
						% (self.__hostname, self.__port, str(e))
			luci_log.info(errstr)
			if LUCI_DEBUG_NET is True:
				luci_log.debug_net_priv('RC:recv2: parsing XML "%s" from %s' \
					% (xml_in, str(e)))
			raise RicciError, errstr

		if not doc or not doc.firstChild:
			if LUCI_DEBUG_MODE is True:
				luci_log.debug_verbose('RC:recv3: empty XML response from %s' \
					% self.__hostname)
			raise RicciError, 'Error an empty response was received from host %s' % self.__hostname

		try:
			if doc.firstChild.nodeName != 'ricci':
				if LUCI_DEBUG_NET is True:
					luci_log.debug_net_priv('RC:recv4: Expecting "<ricci>" tag, got XML "%s" from %s' % (xml_in, self.__hostname))
				raise Exception, 'Expected first XML child node to be "<ricci>"'
		except Exception, e:
			raise RicciError, 'Invalid XML ricci response from host %s' \
					% self.__hostname
		return doc

def get_ricci_communicator(self, hostname, allowed_systems):
	if not hostname:
		return None

	if not self.access_to_host_allowed(hostname, allowed_systems):
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('GRC0: access to host %s is not allowed' % hostname)
		return None

	try:
		rc = RicciCommunicator(hostname)
		if not rc:
			raise Exception, 'initialization failed'
		return rc
	except Exception, e:
		luci_log.info('Error creating a ricci connection to %s: %r %s' \
			% (hostname, e, str(e)))
		return None

def ricci_get_called_hostname(self, ricci):
	return ricci.hostname()
def ricci_get_reported_hostname(self, ricci):
	return ricci.system_name()
def ricci_get_os(self, ricci):
	return ricci.os()
def ricci_get_dom0(self, ricci):
	return ricci.dom0()
def ricci_get_cluster_info(self, ricci):
	return ricci.cluster_info()
def ricci_get_authenticated(self, ricci):
	return ricci.authed()
def ricci_authenticate(self, ricci, password):
	return ricci.auth(password)
def ricci_unauthenticate(self, ricci):
	return ricci.unauth()

########## helpers to process batch as returned by ricci #############

# check the status of batch
# returns (int num, int total)
# * total:
#		total number of modules in batch
# * num:
#		if num == total:
#			 all modules in the batch completed successfuly
#		if num > 0:
#			 last seq. number of module that successfuly finished
#		if num < 0:
#			 module (-num) failed (next module won't be processed)

def batch_status(batch_xml):
	if batch_xml.nodeName != 'batch':
		if LUCI_DEBUG_NET is True:
			try:
				luci_log.debug_net_priv('RC:BS0: batch node. Got \"%s\"' \
					% batch_xml.toxml())
			except:
				pass
		raise RicciError, 'Not an XML batch node'

	total = 0
	last  = 0
	for node in batch_xml.childNodes:
		if node.nodeType == xml.dom.Node.ELEMENT_NODE:
			if node.nodeName == 'module':
				total = total + 1
				status = node.getAttribute('status')
				if status == '0':
					# success
					last = last + 1
				elif status == '3' or status == '4':
					# failure
					last = last + 1
					last = last - 2 * last
	try:
		if LUCI_DEBUG_NET is True:
			luci_log.debug_net_priv('RC:BS1: (%d, %d) for batch_status("%s")' \
				% (last, total, batch_xml.toxml()))
	except Exception, e:
		if LUCI_DEBUG_MODE is True:
			luci_log.debug_verbose('RC:BS2: error %r %s: (%d, %d)' \
				% (e, str(e), last, total))
	return (last, total)

# extract error_code from module's response
# * module_num:
#			1-based seq. number of module to process
#
# returns (int error_code, string error_msg)
# * error_code: each module defines own error codes, which are >0
#		-101 - in progress
#		-102 - scheduled
#		-103 - removed from schedule
#		-104 - failed to execute module
#
#		>-3 - module executed. Following codes are defined:
#		 -2 - API error
#		 -1 - undefined error occured (msg not necesarily very informative)
#		  0 - no error (msg is empty string)
#		 >0 - predefined error has occured
#						(check respective API, msg will be fully descriptive)
# * error_msg:  error message

def extract_module_status(batch_xml, module_num=1):
	if batch_xml.nodeName != 'batch':
		if LUCI_DEBUG_NET is True:
			luci_log.debug_net_priv('RC:EMS0: Expecting "batch" got "%s"' \
				% batch_xml.toxml())
		raise RicciError, 'Invalid XML node; expecting a batch node'

	c = 0
	for node in batch_xml.childNodes:
		if node.nodeType == xml.dom.Node.ELEMENT_NODE:
			if node.nodeName == 'module':
				module_xml = node
				c = c + 1
				if c == module_num:
					status = module_xml.getAttribute('status')
					if status == '0' or status == '4':
						# module executed, dig deeper into request
						for node_i in module_xml.childNodes:
							if node_i.nodeType == xml.dom.Node.ELEMENT_NODE:
								if node_i.nodeName == 'API_error':
									return -2, 'API error'
								elif node_i.nodeName == 'response':
									for node_j in node_i.childNodes:
										if node_j.nodeType == xml.dom.Node.ELEMENT_NODE:
											if node_j.nodeName == 'function_response':
												code = -11111111
												msg = 'BUG'
												for var in node_j.childNodes:
													if var.nodeType == xml.dom.Node.ELEMENT_NODE:
														if var.nodeName == 'var':
															if var.getAttribute('name') == 'success' and var.getAttribute('value') == 'true':
																return 0, ''
															elif var.getAttribute('name') == 'error_code':
																code = int(var.getAttribute('value'))
															elif var.getAttribute('name') == 'error_description':
																msg = var.getAttribute('value')
												return code, msg

					elif status == '1':
						return -102, 'module scheduled for execution'
					elif status == '2':
						return -101, 'module is being executed'
					elif status == '3':
						return -104, 'failed to locate/execute module'
					elif status == '5':
						return -103, 'module removed from schedule'

	raise RicciError, 'no %dth module in the batch, or malformed response' % module_num
